"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ChainConnectorPlugin = void 0;
const lisk_sdk_1 = require("lisk-sdk");
const active_validators_update_1 = require("./active_validators_update");
const certificate_generation_1 = require("./certificate_generation");
const constants_1 = require("./constants");
const db_1 = require("./db");
const endpoint_1 = require("./endpoint");
const schemas_1 = require("./schemas");
const inbox_update_1 = require("./inbox_update");
const utils_1 = require("./utils");
const { address, ed } = lisk_sdk_1.cryptography;
class ChainConnectorPlugin extends lisk_sdk_1.BasePlugin {
    constructor() {
        super(...arguments);
        this.endpoint = new endpoint_1.Endpoint();
        this.configSchema = schemas_1.configSchema;
    }
    get nodeModulePath() {
        return __filename;
    }
    async init(context) {
        var _a;
        await super.init(context);
        this._ccuFrequency = (_a = this.config.ccuFrequency) !== null && _a !== void 0 ? _a : constants_1.CCU_FREQUENCY;
        if (this.config.maxCCUSize > constants_1.CCU_TOTAL_CCM_SIZE) {
            throw new Error(`maxCCUSize cannot be greater than ${constants_1.CCU_TOTAL_CCM_SIZE} bytes.`);
        }
        this._receivingChainID = Buffer.from(this.config.receivingChainID, 'hex');
        this._maxCCUSize = this.config.maxCCUSize;
        this._isSaveCCU = this.config.isSaveCCU;
        this._registrationHeight = this.config.registrationHeight;
        this._ccuSaveLimit = this.config.ccuSaveLimit;
        this._receivingChainFinalizedHeight = 0;
        this._heightToDeleteIndex = new Map();
    }
    async load() {
        this._chainConnectorPluginDB = await (0, db_1.getDBInstance)(this.dataPath);
        this._chainConnectorStore = new db_1.ChainConnectorStore(this._chainConnectorPluginDB);
        this.endpoint.load(this.config, this._chainConnectorStore);
        this._sendingChainClient = this.apiClient;
        this._ownChainID = Buffer.from(this.appConfig.genesis.chainID, 'hex');
        if (this._receivingChainID[0] !== this._ownChainID[0]) {
            throw new Error('Receiving Chain ID network does not match the sending chain network.');
        }
        this._isReceivingChainMainchain = !(0, utils_1.getMainchainID)(this._ownChainID).equals(this._ownChainID);
        this._sendingChainClient.subscribe('chain_newBlock', async (data) => this._newBlockHandler(data));
        this._sendingChainClient.subscribe('chain_deleteBlock', async (data) => this._deleteBlockHandler(data));
    }
    async unload() {
        if (this._receivingChainClient) {
            await this._receivingChainClient.disconnect();
        }
        if (this._sendingChainClient) {
            await this._sendingChainClient.disconnect();
        }
        this._chainConnectorStore.close();
    }
    async _newBlockReceivingChainHandler(_) {
        try {
            const { finalizedHeight } = await this._receivingChainClient.invoke('system_getNodeInfo');
            this._receivingChainFinalizedHeight = finalizedHeight;
            const { inbox } = await this._receivingChainClient.invoke('interoperability_getChannel', { chainID: this._ownChainID.toString('hex') });
            if (!inbox) {
                throw new Error('No channel data available on receiving chain.');
            }
            const { lastCertificate } = await this._receivingChainClient.invoke('interoperability_getChainAccount', { chainID: this._ownChainID.toString('hex') });
            if (!lastCertificate) {
                throw new Error('No chain data available on receiving chain.');
            }
            this._heightToDeleteIndex.set(finalizedHeight, {
                inboxSize: inbox.size,
                lastCertificateHeight: lastCertificate.height,
            });
        }
        catch (error) {
            this.logger.debug(error, 'No Channel or Chain Data: Sending chain is not registered yet on receiving chain.');
        }
        await this._cleanup();
    }
    async _newBlockHandler(data) {
        const { blockHeader: receivedBlock } = data;
        const newBlockHeader = lisk_sdk_1.chain.BlockHeader.fromJSON(receivedBlock).toObject();
        let chainAccountJSON;
        try {
            try {
                chainAccountJSON = await this._receivingChainClient.invoke('interoperability_getChainAccount', { chainID: this._ownChainID.toString('hex') });
                if (!chainAccountJSON || (chainAccountJSON && !chainAccountJSON.lastCertificate)) {
                    this.logger.info('Sending chain is not registered to the receiving chain yet and has no chain data.');
                    await this._saveDataOnNewBlock(newBlockHeader);
                    return;
                }
            }
            catch (error) {
                await this._saveDataOnNewBlock(newBlockHeader);
                await this._initializeReceivingChainClient();
                this.logger.error(error, 'Error occurred while accessing receivingChainAPIClient but all data is saved on newBlock.');
                return;
            }
            this._lastCertificate = (0, utils_1.chainAccountDataJSONToObj)(chainAccountJSON).lastCertificate;
            const { aggregateCommits, blockHeaders, validatorsHashPreimage, crossChainMessages } = await this._saveDataOnNewBlock(newBlockHeader);
            if (this._ccuFrequency <= newBlockHeader.height - this._lastCertificate.height) {
                const computedCCUParams = await this._computeCCUParams(blockHeaders, aggregateCommits, validatorsHashPreimage, crossChainMessages);
                if (computedCCUParams) {
                    try {
                        await this._submitCCU(lisk_sdk_1.codec.encode(lisk_sdk_1.ccuParamsSchema, computedCCUParams.ccuParams));
                        if (computedCCUParams.lastCCMToBeSent) {
                            await this._chainConnectorStore.setLastSentCCM(computedCCUParams.lastCCMToBeSent);
                        }
                    }
                    catch (error) {
                        this.logger.info({ err: error }, `Error occured while submitting CCU for the blockHeader at height: ${newBlockHeader.height}`);
                        return;
                    }
                }
                else {
                    this.logger.info(`No valid CCU can be generated for the height: ${newBlockHeader.height}`);
                }
            }
        }
        catch (error) {
            this.logger.error(error, 'Failed while handling the new block');
        }
    }
    async _computeCCUParams(blockHeaders, aggregateCommits, validatorsHashPreimage, ccmsFromEvents) {
        var _a;
        const newCertificate = await this._findNextCertificate(aggregateCommits, blockHeaders, validatorsHashPreimage);
        if (!newCertificate && this._lastCertificate.height === 0) {
            return;
        }
        const lastSentCCM = (_a = (await this._chainConnectorStore.getLastSentCCM())) !== null && _a !== void 0 ? _a : {
            nonce: constants_1.DEFAULT_LAST_CCM_SENT_NONCE,
            height: this._lastCertificate.height,
        };
        let activeValidatorsUpdate = {
            blsKeysUpdate: [],
            bftWeightsUpdate: [],
            bftWeightsUpdateBitmap: constants_1.EMPTY_BYTES,
        };
        let certificate = constants_1.EMPTY_BYTES;
        let certificateThreshold;
        let outboxRootWitness;
        const ccmsToBeIncluded = ccmsFromEvents.filter(record => record.height >= lastSentCCM.height &&
            record.height <= (newCertificate ? newCertificate.height : this._lastCertificate.height));
        const channelDataOnReceivingChain = await this._receivingChainClient.invoke('interoperability_getChannel', { chainID: this._ownChainID.toString('hex') });
        if (!(channelDataOnReceivingChain === null || channelDataOnReceivingChain === void 0 ? void 0 : channelDataOnReceivingChain.inbox)) {
            this.logger.info('Receiving chain is not registered yet on the sending chain.');
            return;
        }
        const inboxSizeOnReceivingChain = (0, utils_1.channelDataJSONToObj)(channelDataOnReceivingChain).inbox.size;
        const receivingChainChannelDataJSON = await this._sendingChainClient.invoke('interoperability_getChannel', { chainID: this._receivingChainID.toString('hex') });
        if (!(receivingChainChannelDataJSON === null || receivingChainChannelDataJSON === void 0 ? void 0 : receivingChainChannelDataJSON.outbox)) {
            this.logger.info('Sending chain is not registered yet on the receiving chain.');
            return;
        }
        const outboxSizeOnSendingChain = (0, utils_1.channelDataJSONToObj)(receivingChainChannelDataJSON).outbox
            .size;
        const messageWitnessHashesForCCMs = (0, inbox_update_1.calculateMessageWitnesses)(inboxSizeOnReceivingChain, outboxSizeOnSendingChain, lastSentCCM, ccmsToBeIncluded, this._maxCCUSize);
        const { crossChainMessages, lastCCMToBeSent, messageWitnessHashes } = messageWitnessHashesForCCMs;
        if (!newCertificate) {
            if (crossChainMessages.length === 0) {
                this.logger.info('CCU cant be created as there are no pending CCMs for the last certificate.');
                return;
            }
            outboxRootWitness = {
                bitmap: constants_1.EMPTY_BYTES,
                siblingHashes: [],
            };
            const validatorsDataAtLastCertificate = validatorsHashPreimage.find(validatorsData => validatorsData.validatorsHash.equals(this._lastCertificate.validatorsHash));
            if (!validatorsDataAtLastCertificate) {
                throw new Error('No validatorsData found for the lastCertificate.');
            }
            certificateThreshold = validatorsDataAtLastCertificate.certificateThreshold;
        }
        else {
            if (!this._lastCertificate.validatorsHash.equals(newCertificate.validatorsHash)) {
                const validatorsUpdateResult = (0, active_validators_update_1.calculateActiveValidatorsUpdate)(newCertificate, validatorsHashPreimage, this._lastCertificate);
                activeValidatorsUpdate = validatorsUpdateResult.activeValidatorsUpdate;
                certificateThreshold = validatorsUpdateResult.certificateThreshold;
            }
            else {
                const validatorsDataAtLastCertificate = validatorsHashPreimage.find(validatorsData => validatorsData.validatorsHash.equals(this._lastCertificate.validatorsHash));
                if (!validatorsDataAtLastCertificate) {
                    throw new Error('No validatorsData found for the lastCertificate.');
                }
                certificateThreshold = validatorsDataAtLastCertificate.certificateThreshold;
            }
            const ccmsDataAtCertificateHeight = ccmsToBeIncluded.find(ccmsData => ccmsData.height === newCertificate.height);
            if (crossChainMessages.length === 0) {
                outboxRootWitness = {
                    bitmap: constants_1.EMPTY_BYTES,
                    siblingHashes: [],
                };
            }
            else {
                outboxRootWitness = ccmsDataAtCertificateHeight === null || ccmsDataAtCertificateHeight === void 0 ? void 0 : ccmsDataAtCertificateHeight.inclusionProof;
            }
            certificate = lisk_sdk_1.codec.encode(lisk_sdk_1.certificateSchema, newCertificate);
        }
        return {
            ccuParams: {
                sendingChainID: this._ownChainID,
                activeValidatorsUpdate,
                certificate,
                certificateThreshold,
                inboxUpdate: {
                    crossChainMessages,
                    messageWitnessHashes,
                    outboxRootWitness,
                },
            },
            lastCCMToBeSent,
        };
    }
    async _findNextCertificate(aggregateCommits, blockHeaders, validatorsHashPreimage) {
        if (aggregateCommits.length === 0) {
            return undefined;
        }
        if (this._lastCertificate.height === 0) {
            for (const aggregateCommit of aggregateCommits) {
                if (aggregateCommit.height < this._registrationHeight) {
                    continue;
                }
                const firstCertificate = (0, certificate_generation_1.getCertificateFromAggregateCommit)(aggregateCommit, blockHeaders);
                return firstCertificate;
            }
            return undefined;
        }
        const bftHeights = await this._sendingChainClient.invoke('consensus_getBFTHeights');
        return (0, certificate_generation_1.getNextCertificateFromAggregateCommits)(blockHeaders, aggregateCommits, validatorsHashPreimage, bftHeights, this._lastCertificate);
    }
    async _saveDataOnNewBlock(newBlockHeader) {
        var _a, _b;
        const blockHeaders = await this._chainConnectorStore.getBlockHeaders();
        const blockHeaderIndex = blockHeaders.findIndex(header => header.height === newBlockHeader.height);
        if (blockHeaderIndex > -1) {
            blockHeaders[blockHeaderIndex] = newBlockHeader;
        }
        else {
            blockHeaders.push(newBlockHeader);
        }
        const events = await this._sendingChainClient.invoke('chain_getEvents', { height: newBlockHeader.height });
        const { modules: modulesMetadata } = await this._sendingChainClient.invoke('system_getMetadata');
        const interoperabilityMetadata = modulesMetadata.find(m => m.name === constants_1.MODULE_NAME_INTEROPERABILITY);
        if (!interoperabilityMetadata) {
            throw new Error(`No metadata found for ${constants_1.MODULE_NAME_INTEROPERABILITY} module.`);
        }
        const ccmsFromEvents = [];
        ccmInEventsCheck: if (events && events.length > 0) {
            const ccmSendSuccessEvents = events.filter(eventAttr => eventAttr.name === constants_1.CCM_SEND_SUCCESS && eventAttr.module === constants_1.MODULE_NAME_INTEROPERABILITY);
            const ccmProcessedEvents = events.filter(eventAttr => eventAttr.name === constants_1.CCM_PROCESSED && eventAttr.module === constants_1.MODULE_NAME_INTEROPERABILITY);
            if (ccmSendSuccessEvents.length === 0 && ccmProcessedEvents.length === 0) {
                break ccmInEventsCheck;
            }
            if (ccmSendSuccessEvents.length > 0) {
                const ccmSendSuccessEventInfo = interoperabilityMetadata.events.filter(e => e.name === constants_1.CCM_SEND_SUCCESS);
                if (!((_a = ccmSendSuccessEventInfo === null || ccmSendSuccessEventInfo === void 0 ? void 0 : ccmSendSuccessEventInfo[0]) === null || _a === void 0 ? void 0 : _a.data)) {
                    throw new Error('No schema found for "ccmSendSuccess" event data.');
                }
                for (const e of ccmSendSuccessEvents) {
                    const eventData = lisk_sdk_1.codec.decode(ccmSendSuccessEventInfo[0].data, Buffer.from(e.data, 'hex'));
                    ccmsFromEvents.push(eventData.ccm);
                }
            }
            if (ccmProcessedEvents.length > 0) {
                const ccmProcessedEventInfo = interoperabilityMetadata.events.filter(e => e.name === constants_1.CCM_PROCESSED);
                if (!((_b = ccmProcessedEventInfo === null || ccmProcessedEventInfo === void 0 ? void 0 : ccmProcessedEventInfo[0]) === null || _b === void 0 ? void 0 : _b.data)) {
                    throw new Error('No schema found for "ccmProcessed" event data.');
                }
                for (const e of ccmProcessedEvents) {
                    const eventData = lisk_sdk_1.codec.decode(ccmProcessedEventInfo[0].data, Buffer.from(e.data, 'hex'));
                    if (eventData.result === 1) {
                        ccmsFromEvents.push(eventData.ccm);
                    }
                }
            }
        }
        const store = interoperabilityMetadata.stores.find(s => s.data.$id === '/modules/interoperability/outbox');
        const outboxKey = Buffer.concat([
            Buffer.from(store === null || store === void 0 ? void 0 : store.key, 'hex'),
            lisk_sdk_1.cryptography.utils.hash(this._receivingChainID),
        ]).toString('hex');
        const proveResponseJSON = await this._sendingChainClient.invoke('state_prove', {
            queryKeys: [outboxKey],
        });
        const proveResponseObj = (0, utils_1.proveResponseJSONToObj)(proveResponseJSON);
        const outboxRootWitness = {
            bitmap: proveResponseObj.proof.queries[0].bitmap,
            siblingHashes: proveResponseObj.proof.siblingHashes,
        };
        const crossChainMessages = await this._chainConnectorStore.getCrossChainMessages();
        let receivingChainOutboxSize = 0;
        try {
            const receivingChainChannelDataJSON = await this._sendingChainClient.invoke('interoperability_getChannel', { chainID: this._receivingChainID.toString('hex') });
            receivingChainOutboxSize = receivingChainChannelDataJSON.outbox.size;
        }
        catch (error) {
            this.logger.debug(error, 'No Channel Data: Receiving chain is not registered yet on sending chain');
        }
        crossChainMessages.push({
            ccms: this._isReceivingChainMainchain
                ? ccmsFromEvents
                : ccmsFromEvents.filter(ccm => ccm.receivingChainID.equals(this._receivingChainID)),
            height: newBlockHeader.height,
            inclusionProof: outboxRootWitness,
            outboxSize: receivingChainOutboxSize,
        });
        await this._chainConnectorStore.setCrossChainMessages(crossChainMessages);
        const validatorsHashPreimage = await this._chainConnectorStore.getValidatorsHashPreimage();
        const bftParametersJSON = await this._sendingChainClient.invoke('consensus_getBFTParametersActiveValidators', { height: newBlockHeader.height });
        const bftParametersObj = (0, utils_1.bftParametersJSONToObj)(bftParametersJSON);
        const validatorsDataIndex = validatorsHashPreimage.findIndex(v => v.validatorsHash.equals(bftParametersObj.validatorsHash));
        if (validatorsDataIndex === -1) {
            const activeValidators = bftParametersObj.validators;
            validatorsHashPreimage.push({
                certificateThreshold: bftParametersObj.certificateThreshold,
                validators: activeValidators,
                validatorsHash: bftParametersObj.validatorsHash,
            });
        }
        const aggregateCommits = await this._chainConnectorStore.getAggregateCommits();
        if (!newBlockHeader.aggregateCommit.aggregationBits.equals(constants_1.EMPTY_BYTES) ||
            !newBlockHeader.aggregateCommit.certificateSignature.equals(constants_1.EMPTY_BYTES)) {
            const aggregateCommitIndex = aggregateCommits.findIndex(commit => commit.height === newBlockHeader.aggregateCommit.height);
            if (aggregateCommitIndex > -1) {
                aggregateCommits[aggregateCommitIndex] = newBlockHeader.aggregateCommit;
            }
            else {
                aggregateCommits.push(newBlockHeader.aggregateCommit);
            }
        }
        await this._chainConnectorStore.setBlockHeaders(blockHeaders);
        await this._chainConnectorStore.setAggregateCommits(aggregateCommits);
        await this._chainConnectorStore.setValidatorsHashPreimage(validatorsHashPreimage);
        return {
            blockHeaders,
            aggregateCommits,
            validatorsHashPreimage,
            crossChainMessages,
        };
    }
    async _deleteBlockHandler(data) {
        const { blockHeader: receivedBlock } = data;
        const deletedBlockHeader = lisk_sdk_1.chain.BlockHeader.fromJSON(receivedBlock).toObject();
        const crossChainMessages = await this._chainConnectorStore.getCrossChainMessages();
        const indexForCCMEvents = crossChainMessages.findIndex(ccm => ccm.height === deletedBlockHeader.height);
        crossChainMessages.splice(indexForCCMEvents, 1);
        await this._chainConnectorStore.setCrossChainMessages(crossChainMessages);
        const findIndexByHeight = (someData) => someData.findIndex(datum => datum.height === deletedBlockHeader.height);
        const blockHeaders = await this._chainConnectorStore.getBlockHeaders();
        const blockHeaderIndex = findIndexByHeight(blockHeaders);
        if (blockHeaderIndex !== -1) {
            blockHeaders.splice(blockHeaderIndex, 1);
            await this._chainConnectorStore.setBlockHeaders(blockHeaders);
        }
        if (!deletedBlockHeader.aggregateCommit.aggregationBits.equals(constants_1.EMPTY_BYTES) ||
            !deletedBlockHeader.aggregateCommit.certificateSignature.equals(constants_1.EMPTY_BYTES)) {
            const aggregateCommits = await this._chainConnectorStore.getAggregateCommits();
            const aggregateCommitIndex = aggregateCommits.findIndex(commit => commit.height === deletedBlockHeader.aggregateCommit.height);
            if (aggregateCommitIndex > -1) {
                aggregateCommits.splice(aggregateCommitIndex, 1);
            }
            await this._chainConnectorStore.setAggregateCommits(aggregateCommits);
        }
        const validatorsHashPreimage = await this._chainConnectorStore.getValidatorsHashPreimage();
        const validatorsHashMap = blockHeaders.reduce((prev, curr) => {
            prev[curr.validatorsHash.toString('hex')] = true;
            return prev;
        }, {});
        const updatedValidatorsHashPreimages = validatorsHashPreimage.filter(vhp => validatorsHashMap[vhp.validatorsHash.toString('hex')]);
        if (updatedValidatorsHashPreimages.length !== validatorsHashPreimage.length) {
            await this._chainConnectorStore.setValidatorsHashPreimage(updatedValidatorsHashPreimages);
        }
    }
    async _cleanup() {
        if (this._ccuSaveLimit !== -1) {
            const listOfCCUs = await this._chainConnectorStore.getListOfCCUs();
            if (listOfCCUs.length > this._ccuSaveLimit) {
                await this._chainConnectorStore.setListOfCCUs(listOfCCUs.slice(-this._ccuSaveLimit));
            }
            let finalizedInfoAtHeight = this._heightToDeleteIndex.get(this._receivingChainFinalizedHeight);
            if (!finalizedInfoAtHeight) {
                for (let i = 1; i < this._heightToDeleteIndex.size; i += 1) {
                    if (this._heightToDeleteIndex.get(this._receivingChainFinalizedHeight - i)) {
                        finalizedInfoAtHeight = this._heightToDeleteIndex.get(this._receivingChainFinalizedHeight - i);
                        break;
                    }
                }
            }
            const crossChainMessages = await this._chainConnectorStore.getCrossChainMessages();
            const ccmsAfterLastCertificate = crossChainMessages.filter(ccm => ccm.outboxSize >= (finalizedInfoAtHeight ? finalizedInfoAtHeight.inboxSize : 0));
            await this._chainConnectorStore.setCrossChainMessages(ccmsAfterLastCertificate);
            const blockHeaders = await this._chainConnectorStore.getBlockHeaders();
            const updatedBlockHeaders = blockHeaders.filter(blockHeader => blockHeader.height >=
                (finalizedInfoAtHeight ? finalizedInfoAtHeight.lastCertificateHeight : 0));
            await this._chainConnectorStore.setBlockHeaders(updatedBlockHeaders);
            const aggregateCommits = await this._chainConnectorStore.getAggregateCommits();
            await this._chainConnectorStore.setAggregateCommits(aggregateCommits.filter(aggregateCommit => aggregateCommit.height >=
                (finalizedInfoAtHeight ? finalizedInfoAtHeight.lastCertificateHeight : 0)));
            const validatorsHashPreimage = await this._chainConnectorStore.getValidatorsHashPreimage();
            const validatorsHashMap = updatedBlockHeaders.reduce((prev, curr) => {
                prev[curr.validatorsHash.toString('hex')] = true;
                return prev;
            }, {});
            const updatedValidatorsHashPreimages = validatorsHashPreimage.filter(vhp => validatorsHashMap[vhp.validatorsHash.toString('hex')]);
            if (updatedValidatorsHashPreimages.length !== validatorsHashPreimage.length) {
                await this._chainConnectorStore.setValidatorsHashPreimage(updatedValidatorsHashPreimages);
            }
            this._heightToDeleteIndex.forEach((_, key) => {
                if (key < this._receivingChainFinalizedHeight) {
                    this._heightToDeleteIndex.delete(key);
                }
            });
        }
    }
    async _submitCCU(ccuParams) {
        if (!this._chainConnectorStore.privateKey) {
            throw new Error('There is no key enabled to submit CCU');
        }
        const relayerPublicKey = ed.getPublicKeyFromPrivateKey(this._chainConnectorStore.privateKey);
        const targetCommand = this._isReceivingChainMainchain
            ? constants_1.COMMAND_NAME_SUBMIT_MAINCHAIN_CCU
            : constants_1.COMMAND_NAME_SUBMIT_SIDECHAIN_CCU;
        const { nonce } = await this._receivingChainClient.invoke('auth_getAuthAccount', {
            address: address.getLisk32AddressFromPublicKey(relayerPublicKey),
        });
        const { chainID: chainIDStr } = await this._receivingChainClient.invoke('system_getNodeInfo');
        const chainID = Buffer.from(chainIDStr, 'hex');
        const tx = new lisk_sdk_1.Transaction({
            module: constants_1.MODULE_NAME_INTEROPERABILITY,
            command: targetCommand,
            nonce: BigInt(nonce),
            senderPublicKey: relayerPublicKey,
            fee: BigInt(this.config.ccuFee),
            params: ccuParams,
            signatures: [],
        });
        tx.sign(chainID, this._chainConnectorStore.privateKey);
        let result;
        if (this._isSaveCCU) {
            result = { transactionId: tx.id.toString('hex') };
        }
        else {
            result = await this._receivingChainClient.invoke('txpool_postTransaction', {
                transaction: tx.getBytes().toString('hex'),
            });
        }
        const listOfCCUs = await this._chainConnectorStore.getListOfCCUs();
        listOfCCUs.push(tx.toObject());
        await this._chainConnectorStore.setListOfCCUs(listOfCCUs);
        this.logger.info({ transactionID: result.transactionId }, 'Sent CCU transaction');
    }
    async _initializeReceivingChainClient() {
        if (!this.config.receivingChainIPCPath && !this.config.receivingChainWsURL) {
            throw new Error('IPC path and WS url are undefined in the configuration.');
        }
        try {
            if (this.config.receivingChainIPCPath) {
                this._receivingChainClient = await lisk_sdk_1.apiClient.createIPCClient(this.config.receivingChainIPCPath);
            }
            else if (this.config.receivingChainWsURL) {
                this._receivingChainClient = await lisk_sdk_1.apiClient.createWSClient(this.config.receivingChainWsURL);
            }
            this._receivingChainClient.subscribe('chain_newBlock', async (data) => this._newBlockReceivingChainHandler(data));
        }
        catch (error) {
            this.logger.error(error, 'Not able to connect to receivingChainAPIClient. Trying again on next new block.');
        }
    }
}
exports.ChainConnectorPlugin = ChainConnectorPlugin;
//# sourceMappingURL=chain_connector_plugin.js.map