"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ForgerPlugin = void 0;
const lisk_sdk_1 = require("lisk-sdk");
const db_1 = require("./db");
const endpoint_1 = require("./endpoint");
const BLOCKS_BATCH_TO_SYNC = 1000;
const MODULE_POS = 'pos';
const COMMAND_STAKE = 'stake';
const getBinaryAddress = (hexAddressStr) => Buffer.from(hexAddressStr, 'hex').toString('binary');
const getAddressBuffer = (hexAddressStr) => Buffer.from(hexAddressStr, 'hex');
class ForgerPlugin extends lisk_sdk_1.BasePlugin {
    constructor() {
        super(...arguments);
        this.endpoint = new endpoint_1.Endpoint();
    }
    get nodeModulePath() {
        return __filename;
    }
    get events() {
        return ['block:created', 'block:missed'];
    }
    async init(context) {
        await super.init(context);
        this.endpoint.init(this._forgerPluginDB, this.apiClient);
    }
    async load() {
        this._forgerPluginDB = await (0, db_1.getDBInstance)(this.dataPath);
        await this._setForgersList();
        await this._syncForgerInfo();
        this._subscribeToChannel();
    }
    async unload() {
        this._forgerPluginDB.close();
    }
    async _setForgersList() {
        this._forgersList = new lisk_sdk_1.utils.dataStructures.BufferMap();
        const { status: forgersList } = await this.apiClient.invoke('generator_getStatus');
        for (const { address, forging } of forgersList) {
            this._forgersList.set(Buffer.from(address, 'hex'), forging);
        }
    }
    _getForgerHeaderAndTransactionsInfo(header, transactions) {
        const forgerAddress = header.generatorAddress;
        const forgerAddressBuffer = getAddressBuffer(forgerAddress);
        const forgerAddressBinary = getBinaryAddress(forgerAddress);
        return {
            forgerAddress,
            forgerAddressBuffer,
            forgerAddressBinary,
            header,
            transactions,
        };
    }
    async _syncForgerInfo() {
        const lastBlock = await this.apiClient.invoke('chain_getLastBlock');
        const { header: { height: lastBlockHeight }, } = lastBlock;
        const { syncUptoHeight } = await (0, db_1.getForgerSyncInfo)(this._forgerPluginDB);
        const { genesisHeight } = await this.apiClient.invoke('system_getNodeInfo');
        const forgerPluginSyncedHeight = syncUptoHeight === 0 ? genesisHeight : syncUptoHeight;
        if (forgerPluginSyncedHeight === lastBlockHeight) {
            return;
        }
        let needleHeight;
        if (forgerPluginSyncedHeight > lastBlockHeight) {
            await this._forgerPluginDB.clear();
            needleHeight = genesisHeight + 1;
        }
        else {
            needleHeight = forgerPluginSyncedHeight + 1;
        }
        while (needleHeight <= lastBlockHeight) {
            const toHeight = needleHeight +
                (needleHeight + BLOCKS_BATCH_TO_SYNC <= lastBlockHeight
                    ? BLOCKS_BATCH_TO_SYNC
                    : lastBlockHeight - needleHeight);
            const blocks = await this.apiClient.invoke('chain_getBlocksByHeightBetween', {
                from: needleHeight,
                to: toHeight,
            });
            for (const block of blocks.reverse()) {
                const forgerTransactionsInfo = this._getForgerHeaderAndTransactionsInfo(block.header, block.transactions);
                await this._addForgerInfo(block.header, forgerTransactionsInfo);
            }
            needleHeight = toHeight + 1;
        }
        await (0, db_1.setForgerSyncInfo)(this._forgerPluginDB, lastBlockHeight);
        await this._syncForgerInfo();
    }
    _subscribeToChannel() {
        this.apiClient.subscribe('chain_newBlock', async (data) => {
            const { blockHeader } = data;
            const transactions = await this.apiClient.invoke('chain_getTransactionsByHeight', { height: blockHeader.height });
            const forgerTransactionsInfo = this._getForgerHeaderAndTransactionsInfo(blockHeader, transactions);
            const { header: { height }, } = forgerTransactionsInfo;
            await this._addForgerInfo(blockHeader, forgerTransactionsInfo);
            await (0, db_1.setForgerSyncInfo)(this._forgerPluginDB, height);
        });
        this.apiClient.subscribe('chain_deleteBlock', async (data) => {
            const { blockHeader } = data;
            const transactions = await this.apiClient.invoke('chain_getTransactionsByHeight', { height: blockHeader.height });
            const forgerTransactionsInfo = this._getForgerHeaderAndTransactionsInfo(blockHeader, transactions);
            const { header: { height }, } = forgerTransactionsInfo;
            await this._revertForgerInfo(forgerTransactionsInfo);
            await (0, db_1.setForgerSyncInfo)(this._forgerPluginDB, height);
        });
    }
    async _addForgerInfo(header, forgerTransactionsInfo) {
        const { forgerAddressBuffer, forgerAddressBinary, transactions } = forgerTransactionsInfo;
        const forgerInfo = await (0, db_1.getForgerInfo)(this._forgerPluginDB, forgerAddressBinary);
        if (this._forgersList.has(forgerAddressBuffer)) {
            forgerInfo.totalProducedBlocks += 1;
            forgerInfo.totalReceivedFees += this._getFee(transactions);
            await (0, db_1.setForgerInfo)(this._forgerPluginDB, forgerAddressBinary, { ...forgerInfo });
        }
        await this._addStakesReceived(transactions);
        await this._updateMissedBlock(header, transactions);
    }
    async _revertForgerInfo(forgerTransactionsInfo) {
        const { forgerAddressBuffer, forgerAddressBinary, transactions } = forgerTransactionsInfo;
        const forgerInfo = await (0, db_1.getForgerInfo)(this._forgerPluginDB, forgerAddressBinary);
        if (this._forgersList.has(forgerAddressBuffer)) {
            forgerInfo.totalProducedBlocks -= 1;
            forgerInfo.totalReceivedFees -= this._getFee(transactions);
            await (0, db_1.setForgerInfo)(this._forgerPluginDB, forgerAddressBinary, { ...forgerInfo });
        }
        await this._revertStakesReceived(transactions);
    }
    _getForgerReceivedStakes(transactions) {
        const forgerReceivedStakes = {};
        const posModuleMeta = this.apiClient.metadata.find(c => c.name === MODULE_POS);
        if (!posModuleMeta) {
            throw new Error('PoS stakes command is not registered.');
        }
        const voteCommandMeta = posModuleMeta.commands.find(c => c.name === COMMAND_STAKE);
        if (!(voteCommandMeta === null || voteCommandMeta === void 0 ? void 0 : voteCommandMeta.params)) {
            throw new Error('PoS stakes command is not registered.');
        }
        for (const trx of transactions) {
            if (trx.module === MODULE_POS && trx.command === COMMAND_STAKE) {
                const params = lisk_sdk_1.codec.decode(voteCommandMeta.params, Buffer.from(trx.params, 'hex'));
                params.stakes.reduce((acc, curr) => {
                    if (this._forgersList.has(getAddressBuffer(curr.validatorAddress)) &&
                        acc[curr.validatorAddress]) {
                        acc[curr.validatorAddress].amount += BigInt(curr.amount);
                    }
                    else {
                        acc[curr.validatorAddress] = {
                            address: lisk_sdk_1.cryptography.address.getAddressFromPublicKey(Buffer.from(trx.senderPublicKey, 'hex')),
                            amount: BigInt(curr.amount),
                        };
                    }
                    return acc;
                }, forgerReceivedStakes);
            }
        }
        return forgerReceivedStakes;
    }
    async _addStakesReceived(transactions) {
        const forgerReceivedStakes = this._getForgerReceivedStakes(transactions);
        for (const [validatorAddress, stakeReceived] of Object.entries(forgerReceivedStakes)) {
            const forgerInfo = await (0, db_1.getForgerInfo)(this._forgerPluginDB, getBinaryAddress(validatorAddress));
            const stakerIndex = forgerInfo.stakeReceived.findIndex(aStake => aStake.address.equals(stakeReceived.address));
            if (stakerIndex === -1) {
                forgerInfo.stakeReceived.push(stakeReceived);
            }
            else {
                forgerInfo.stakeReceived[stakerIndex].amount += stakeReceived.amount;
                if (forgerInfo.stakeReceived[stakerIndex].amount === BigInt(0)) {
                    forgerInfo.stakeReceived.splice(stakerIndex, 1);
                }
            }
            await (0, db_1.setForgerInfo)(this._forgerPluginDB, getBinaryAddress(validatorAddress), forgerInfo);
        }
    }
    async _revertStakesReceived(transactions) {
        const forgerReceivedStakes = this._getForgerReceivedStakes(transactions);
        for (const [validatorAddress, stakeReceived] of Object.entries(forgerReceivedStakes)) {
            const forgerInfo = await (0, db_1.getForgerInfo)(this._forgerPluginDB, getBinaryAddress(validatorAddress));
            const stakerIndex = forgerInfo.stakeReceived.findIndex(aStake => aStake.address.equals(stakeReceived.address));
            if (stakerIndex !== -1) {
                forgerInfo.stakeReceived[stakerIndex].amount -= BigInt(stakeReceived.amount);
                if (forgerInfo.stakeReceived[stakerIndex].amount === BigInt(0)) {
                    forgerInfo.stakeReceived.splice(stakerIndex, 1);
                }
                await (0, db_1.setForgerInfo)(this._forgerPluginDB, getBinaryAddress(validatorAddress), forgerInfo);
            }
        }
    }
    _getFee(transactions) {
        let fee = BigInt(0);
        for (const txJSON of transactions) {
            const trx = lisk_sdk_1.chain.Transaction.fromJSON(txJSON);
            fee += BigInt(trx.fee);
        }
        return fee;
    }
    async _updateMissedBlock(header, transactions) {
        const { header: { height, timestamp }, forgerAddress, } = this._getForgerHeaderAndTransactionsInfo(header, transactions);
        const previousBlock = await this.apiClient.invoke('chain_getBlockByHeight', {
            height: height - 1,
        });
        const { genesis: { blockTime }, } = await this.apiClient.invoke('system_getNodeInfo');
        const missedBlocks = Math.ceil((timestamp - previousBlock.header.timestamp) / blockTime) - 1;
        if (missedBlocks > 0) {
            const { list: forgersInfo } = await this.apiClient.invoke('chain_getGeneratorList');
            const forgersRoundLength = forgersInfo.length;
            const forgerIndex = forgersInfo.findIndex(f => f.address === forgerAddress);
            const missedBlocksByAddress = {};
            for (let index = 0; index < missedBlocks; index += 1) {
                const rawIndex = (forgerIndex - 1 - index) % forgersRoundLength;
                const forgerRoundIndex = rawIndex >= 0 ? rawIndex : rawIndex + forgersRoundLength;
                const missedForgerInfo = forgersInfo[forgerRoundIndex];
                missedBlocksByAddress[missedForgerInfo.address] =
                    missedBlocksByAddress[missedForgerInfo.address] === undefined
                        ? 1
                        : (missedBlocksByAddress[missedForgerInfo.address] += 1);
            }
        }
    }
}
exports.ForgerPlugin = ForgerPlugin;
//# sourceMappingURL=forger_plugin.js.map