"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.clearBlockHeaders = exports.getContradictingBlockHeader = exports.saveBlockHeaders = exports.getBlockHeaders = exports.getDBInstance = exports.blockHeadersSchema = void 0;
const os = require("os");
const path_1 = require("path");
const fs_extra_1 = require("fs-extra");
const lisk_sdk_1 = require("lisk-sdk");
const { BlockHeader } = lisk_sdk_1.chain;
const { utils } = lisk_sdk_1.cryptography;
exports.blockHeadersSchema = {
    $id: '/lisk/reportMisbehavior/blockHeaders',
    type: 'object',
    required: ['blockHeaders'],
    properties: {
        blockHeaders: {
            type: 'array',
            fieldNumber: 1,
            items: {
                dataType: 'bytes',
            },
        },
    },
};
const getDBInstance = async (dataPath, dbName = 'lisk-framework-report-misbehavior-plugin.db') => {
    const dirPath = (0, path_1.join)(dataPath.replace('~', os.homedir()), 'plugins/data', dbName);
    await (0, fs_extra_1.ensureDir)(dirPath);
    return new lisk_sdk_1.db.Database(dirPath);
};
exports.getDBInstance = getDBInstance;
const getBlockHeaders = async (db, dbKeyBlockHeader) => {
    try {
        const encodedBlockHeaders = await db.get(dbKeyBlockHeader);
        return lisk_sdk_1.codec.decode(exports.blockHeadersSchema, encodedBlockHeaders);
    }
    catch (error) {
        return { blockHeaders: [] };
    }
};
exports.getBlockHeaders = getBlockHeaders;
const saveBlockHeaders = async (db, headerBytes) => {
    const header = BlockHeader.fromBytes(headerBytes);
    const heightBytes = Buffer.alloc(4);
    heightBytes.writeUInt32BE(header.height, 0);
    const dbKey = Buffer.concat([header.generatorAddress, Buffer.from(':', 'utf8'), heightBytes]);
    const { blockHeaders } = await (0, exports.getBlockHeaders)(db, dbKey);
    if (!blockHeaders.find(blockHeader => utils.hash(blockHeader).equals(header.id))) {
        await db.set(dbKey, lisk_sdk_1.codec.encode(exports.blockHeadersSchema, {
            blockHeaders: [...blockHeaders, header.getBytes()],
        }));
        return true;
    }
    return false;
};
exports.saveBlockHeaders = saveBlockHeaders;
const getContradictingBlockHeader = async (db, blockHeader) => {
    const existingHeaders = await new Promise((resolve, reject) => {
        const stream = db.createReadStream({
            gte: Buffer.concat([blockHeader.generatorAddress, Buffer.alloc(4, 0)]),
            lte: Buffer.concat([blockHeader.generatorAddress, Buffer.alloc(4, 255)]),
        });
        const results = [];
        stream
            .on('data', ({ value }) => {
            const { blockHeaders } = lisk_sdk_1.codec.decode(exports.blockHeadersSchema, value);
            for (const encodedHeader of blockHeaders) {
                results.push(encodedHeader.toString('hex'));
            }
        })
            .on('error', error => {
            reject(error);
        })
            .on('end', () => {
            resolve(results);
        });
    });
    for (const headerString of existingHeaders) {
        const header = BlockHeader.fromBytes(Buffer.from(headerString, 'hex'));
        if (blockHeader.id.equals(header.id)) {
            continue;
        }
        const isContradicting = (0, lisk_sdk_1.areDistinctHeadersContradicting)(header, blockHeader);
        if (isContradicting) {
            return header;
        }
    }
    return undefined;
};
exports.getContradictingBlockHeader = getContradictingBlockHeader;
const clearBlockHeaders = async (db, currentHeight) => {
    const keys = await new Promise((resolve, reject) => {
        const stream = db.createReadStream();
        const res = [];
        stream
            .on('data', ({ key, value }) => {
            const { blockHeaders } = lisk_sdk_1.codec.decode(exports.blockHeadersSchema, value);
            for (const encodedHeader of blockHeaders) {
                const decodedBlockHeader = BlockHeader.fromBytes(encodedHeader);
                if (decodedBlockHeader.height < currentHeight - 260000) {
                    res.push(key);
                }
            }
        })
            .on('error', error => {
            reject(error);
        })
            .on('end', () => {
            resolve(res);
        });
    });
    const batch = new lisk_sdk_1.db.Batch();
    for (const k of keys) {
        batch.del(k);
    }
    await db.write(batch);
};
exports.clearBlockHeaders = clearBlockHeaders;
//# sourceMappingURL=db.js.map