"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ReportMisbehaviorPlugin = void 0;
const lisk_sdk_1 = require("lisk-sdk");
const db_1 = require("./db");
const schemas_1 = require("./schemas");
const endpoint_1 = require("./endpoint");
const { address, ed } = lisk_sdk_1.cryptography;
const { BlockHeader, Transaction, TAG_TRANSACTION } = lisk_sdk_1.chain;
class ReportMisbehaviorPlugin extends lisk_sdk_1.BasePlugin {
    constructor() {
        super(...arguments);
        this.configSchema = schemas_1.configSchema;
        this.endpoint = new endpoint_1.Endpoint();
        this._state = { currentHeight: 0 };
    }
    get nodeModulePath() {
        return __filename;
    }
    async init(context) {
        await super.init(context);
        this.endpoint.init(this._state, this.config);
    }
    async load() {
        this._pluginDB = await (0, db_1.getDBInstance)(this.dataPath);
        this._subscribeToChannel();
        this._clearBlockHeadersIntervalId = setInterval(() => {
            (0, db_1.clearBlockHeaders)(this._pluginDB, this._state.currentHeight).catch(error => this.logger.error(error));
        }, this.config.clearBlockHeadersInterval);
    }
    async unload() {
        clearInterval(this._clearBlockHeadersIntervalId);
        this._pluginDB.close();
    }
    _subscribeToChannel() {
        this.apiClient.subscribe('network_newBlock', async (event) => {
            if (!event) {
                this.logger.error('Invalid payload for network_newBlock');
                return;
            }
            const { blockHeader: blockHeaderJSON } = event;
            const headerBytes = lisk_sdk_1.codec.encodeJSON(lisk_sdk_1.blockHeaderSchema, blockHeaderJSON);
            try {
                const saved = await (0, db_1.saveBlockHeaders)(this._pluginDB, headerBytes);
                if (!saved) {
                    return;
                }
                const decodedBlockHeader = BlockHeader.fromBytes(headerBytes);
                if (decodedBlockHeader.height > this._state.currentHeight) {
                    this._state.currentHeight = decodedBlockHeader.height;
                }
                const contradictingBlock = await (0, db_1.getContradictingBlockHeader)(this._pluginDB, decodedBlockHeader);
                if (contradictingBlock && this._state.privateKey) {
                    const encodedTransaction = await this._createPoMTransaction(decodedBlockHeader, contradictingBlock);
                    const result = await this.apiClient.invoke('txpool_postTransaction', {
                        transaction: encodedTransaction,
                    });
                    this.logger.debug('Sent Report misbehavior transaction', result.transactionId);
                }
            }
            catch (error) {
                this.logger.error(error);
            }
        });
    }
    async _createPoMTransaction(contradictingBlock, decodedBlockHeader) {
        var _a;
        const posMeta = this.apiClient.metadata.find(m => m.name === 'pos');
        if (!posMeta) {
            throw new Error('PoS module is not registered in the application.');
        }
        const pomParamsInfo = posMeta.commands.find(m => m.name === 'reportMisbehavior');
        if (!(pomParamsInfo === null || pomParamsInfo === void 0 ? void 0 : pomParamsInfo.params)) {
            throw new Error('PoM params schema is not registered in the application.');
        }
        if (!this._state.publicKey || !this._state.privateKey) {
            throw new Error('Key is not registered.');
        }
        const authAccount = await this.apiClient.invoke('auth_getAuthAccount', {
            address: address.getLisk32AddressFromPublicKey(this._state.publicKey),
        });
        const pomTransactionParams = {
            header1: decodedBlockHeader.getBytes(),
            header2: contradictingBlock.getBytes(),
        };
        const { chainID } = await this.apiClient.invoke('system_getNodeInfo');
        const encodedParams = lisk_sdk_1.codec.encode(pomParamsInfo.params, pomTransactionParams);
        const tx = new Transaction({
            module: posMeta.name,
            command: pomParamsInfo.name,
            nonce: BigInt(authAccount.nonce),
            senderPublicKey: (_a = this._state.publicKey) !== null && _a !== void 0 ? _a : ed.getPublicKeyFromPrivateKey(this._state.privateKey),
            fee: BigInt(this.config.fee),
            params: encodedParams,
            signatures: [],
        });
        tx.signatures.push(ed.signData(TAG_TRANSACTION, Buffer.from(chainID, 'hex'), tx.getSigningBytes(), this._state.privateKey));
        return tx.getBytes().toString('hex');
    }
}
exports.ReportMisbehaviorPlugin = ReportMisbehaviorPlugin;
//# sourceMappingURL=report_misbehavior_plugin.js.map