"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.HashOnionCommand = void 0;
const core_1 = require("@oclif/core");
const fs = require("fs-extra");
const path = require("path");
const cryptography = require("@liskhq/lisk-cryptography");
const validator = require("@liskhq/lisk-validator");
const flags_1 = require("../../utils/flags");
const constants_1 = require("../../constants");
class HashOnionCommand extends core_1.Command {
    async run() {
        const { flags: { output, count, distance, pretty }, } = await this.parse(HashOnionCommand);
        if (distance <= 0 || !validator.isValidInteger(distance)) {
            throw new Error('Distance flag must be an integer and greater than 0.');
        }
        if (count <= 0 || !validator.isValidInteger(count)) {
            throw new Error('Count flag must be an integer and greater than 0.');
        }
        if (output) {
            const { dir } = path.parse(output);
            fs.ensureDirSync(dir);
        }
        const seed = cryptography.utils.generateHashOnionSeed();
        const hashBuffers = cryptography.utils.hashOnion(seed, count, distance);
        const hashes = hashBuffers.map(buf => buf.toString('hex'));
        const result = { count, distance, hashes };
        if (output) {
            if (pretty) {
                fs.writeJSONSync(output, result, { spaces: ' ', mode: constants_1.OWNER_READ_WRITE });
            }
            else {
                fs.writeJSONSync(output, result, { mode: constants_1.OWNER_READ_WRITE });
            }
        }
        else {
            this.printJSON(result, pretty);
        }
    }
    printJSON(message, pretty = false) {
        if (pretty) {
            this.log(JSON.stringify(message, undefined, '  '));
        }
        else {
            this.log(JSON.stringify(message));
        }
    }
}
HashOnionCommand.description = 'Create hash onions to be used by the forger.';
HashOnionCommand.examples = [
    'hash-onion --count=1000000 --distance=2000 --pretty',
    'hash-onion --count=1000000 --distance=2000 --output ~/my_onion.json',
];
HashOnionCommand.flags = {
    output: core_1.Flags.string({
        char: 'o',
        description: 'Output file path',
    }),
    count: core_1.Flags.integer({
        char: 'c',
        description: 'Total number of hashes to produce',
        default: 1000000,
    }),
    distance: core_1.Flags.integer({
        char: 'd',
        description: 'Distance between each hash',
        default: 1000,
    }),
    pretty: flags_1.flagsWithParser.pretty,
};
exports.HashOnionCommand = HashOnionCommand;
//# sourceMappingURL=hash-onion.js.map