"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getFileParams = exports.readParamsFile = exports.checkFileExtension = exports.getParamsFromPrompt = exports.prepareQuestions = exports.transformNestedAsset = exports.transformAsset = exports.readStdIn = exports.readFileSource = exports.isFileSource = exports.getPasswordFromPrompt = exports.getPassphraseFromPrompt = void 0;
const path = require("path");
const fs = require("fs");
const inquirer = require("inquirer");
const readline = require("readline");
const error_1 = require("./error");
const capitalise = (text) => `${text.charAt(0).toUpperCase()}${text.slice(1)}`;
const getPromptVerificationFailError = (displayName) => `${capitalise(displayName)} was not successfully repeated.`;
const splitSource = (source) => {
    const delimiter = ':';
    const sourceParts = source.split(delimiter);
    return {
        sourceType: sourceParts[0],
        sourceIdentifier: sourceParts.slice(1).join(delimiter),
    };
};
const getPassphraseFromPrompt = async (displayName = 'passphrase', shouldConfirm = false) => {
    const questions = [
        {
            type: 'password',
            name: 'passphrase',
            message: `Please enter ${displayName}: `,
        },
    ];
    if (shouldConfirm) {
        questions.push({
            type: 'password',
            name: 'passphraseRepeat',
            message: `Please re-enter ${displayName}: `,
        });
    }
    const { passphrase, passphraseRepeat } = await inquirer.prompt(questions);
    if (!passphrase || (shouldConfirm && passphrase !== passphraseRepeat)) {
        throw new error_1.ValidationError(getPromptVerificationFailError(displayName));
    }
    return passphrase;
};
exports.getPassphraseFromPrompt = getPassphraseFromPrompt;
const getPasswordFromPrompt = async (displayName = 'password', shouldConfirm = false) => {
    const questions = [
        {
            type: 'password',
            name: 'password',
            message: `Please enter ${displayName}: `,
        },
    ];
    if (shouldConfirm) {
        questions.push({
            type: 'password',
            name: 'passwordRepeat',
            message: `Please re-enter ${displayName}: `,
        });
    }
    const { password, passwordRepeat } = await inquirer.prompt(questions);
    if (!password || (shouldConfirm && password !== passwordRepeat)) {
        throw new error_1.ValidationError(getPromptVerificationFailError(displayName));
    }
    return password;
};
exports.getPasswordFromPrompt = getPasswordFromPrompt;
const getFileDoesNotExistError = (filePath) => `File at ${filePath} does not exist.`;
const getFileUnreadableError = (filePath) => `File at ${filePath} could not be read.`;
const getDataFromFile = (filePath) => fs.readFileSync(filePath, 'utf8');
const ERROR_DATA_MISSING = 'No data was provided.';
const ERROR_DATA_SOURCE = 'Unknown data source type.';
const INVALID_JSON_FILE = 'Not a JSON file.';
const FILE_NOT_FOUND = 'No such file or directory.';
const isFileSource = (source) => {
    if (!source) {
        return false;
    }
    const delimiter = ':';
    const sourceParts = source.split(delimiter);
    if (sourceParts.length === 2 && sourceParts[0] === 'file') {
        return true;
    }
    return false;
};
exports.isFileSource = isFileSource;
const readFileSource = async (source) => {
    if (!source) {
        throw new error_1.ValidationError(ERROR_DATA_MISSING);
    }
    const { sourceType, sourceIdentifier: filePath } = splitSource(source);
    if (sourceType !== 'file') {
        throw new error_1.ValidationError(ERROR_DATA_SOURCE);
    }
    try {
        return getDataFromFile(filePath);
    }
    catch (error) {
        const { message } = error;
        if (message.match(/ENOENT/)) {
            throw new error_1.FileSystemError(getFileDoesNotExistError(filePath));
        }
        if (message.match(/EACCES/)) {
            throw new error_1.FileSystemError(getFileUnreadableError(filePath));
        }
        throw error;
    }
};
exports.readFileSource = readFileSource;
const DEFAULT_TIMEOUT = 100;
const readStdIn = async () => {
    const readFromStd = new Promise((resolve, reject) => {
        const lines = [];
        const rl = readline.createInterface({ input: process.stdin });
        setTimeout(() => {
            reject(new Error(`Timed out after ${DEFAULT_TIMEOUT} ms`));
        }, DEFAULT_TIMEOUT);
        const handleClose = () => {
            resolve(lines);
        };
        return rl.on('line', line => lines.push(line)).on('close', handleClose);
    });
    return readFromStd;
};
exports.readStdIn = readStdIn;
const getNestedPropertyTemplate = (schema) => {
    const keyValEntries = Object.entries(schema.properties);
    const template = {};
    for (let i = 0; i < keyValEntries.length; i += 1) {
        const [schemaPropertyName, schemaPropertyValue] = keyValEntries[i];
        if (schemaPropertyValue.type === 'array') {
            if (schemaPropertyValue.items.type === 'object') {
                template[schemaPropertyName] = Object.keys(schemaPropertyValue.items.properties);
            }
        }
    }
    return template;
};
const castValue = (val, schemaType) => {
    if (schemaType === 'object') {
        return JSON.parse(val);
    }
    if (schemaType === 'array') {
        return val !== '' ? val.split(',') : [];
    }
    if (schemaType === 'uint64' || schemaType === 'sint64') {
        return BigInt(val);
    }
    if (schemaType === 'uint32' || schemaType === 'sint32') {
        return Number(val);
    }
    return val;
};
const transformAsset = (schema, data) => {
    const propertySchema = Object.values(schema.properties);
    const assetData = {};
    return Object.entries(data).reduce((acc, curr, index) => {
        const propSchema = propertySchema[index];
        const schemaType = propSchema.type || propSchema.dataType;
        acc[curr[0]] = castValue(curr[1], schemaType);
        return acc;
    }, assetData);
};
exports.transformAsset = transformAsset;
const transformNestedAsset = (schema, data) => {
    const template = getNestedPropertyTemplate(schema);
    const result = {};
    const items = [];
    for (const assetData of data) {
        const [[key, val]] = Object.entries(assetData);
        const templateValues = template[key];
        const initData = {};
        const valObject = val.split(',').reduce((acc, curr, index) => {
            acc[templateValues[index]] = Number.isInteger(Number(curr)) ? Number(curr) : curr;
            return acc;
        }, initData);
        items.push(valObject);
        result[key] = items;
    }
    return result;
};
exports.transformNestedAsset = transformNestedAsset;
const prepareQuestions = (schema) => {
    const keyValEntries = Object.entries(schema.properties);
    const questions = [];
    for (const [schemaPropertyName, schemaPropertyValue] of keyValEntries) {
        if (schemaPropertyValue.type === 'array') {
            let commaSeparatedKeys = [];
            if (schemaPropertyValue.items.type === 'object') {
                commaSeparatedKeys = Object.keys(schemaPropertyValue.items.properties);
            }
            questions.push({
                type: 'input',
                name: schemaPropertyName,
                message: `Please enter: ${schemaPropertyName}(${commaSeparatedKeys.length ? commaSeparatedKeys.join(', ') : 'comma separated values (a,b)'}): `,
            });
            if (schemaPropertyValue.items.type === 'object') {
                questions.push({
                    type: 'confirm',
                    name: 'askAgain',
                    message: `Want to enter another ${schemaPropertyName}(${commaSeparatedKeys.join(', ')})`,
                });
            }
        }
        else {
            questions.push({
                type: 'input',
                name: schemaPropertyName,
                message: `Please enter: ${schemaPropertyName}: `,
            });
        }
    }
    return questions;
};
exports.prepareQuestions = prepareQuestions;
const getParamsFromPrompt = async (assetSchema, output = []) => {
    const questions = (0, exports.prepareQuestions)(assetSchema);
    if (questions.length === 0) {
        return {};
    }
    let isTypeConfirm = false;
    const result = await inquirer.prompt(questions).then(async (answer) => {
        const inquirerResult = answer;
        isTypeConfirm = typeof inquirerResult.askAgain === 'boolean';
        if (inquirerResult.askAgain) {
            output.push(inquirerResult);
            return (0, exports.getParamsFromPrompt)(assetSchema, output);
        }
        output.push(inquirerResult);
        return Promise.resolve(answer);
    });
    const filteredResult = output.map(({ askAgain, ...assetProps }) => assetProps);
    return isTypeConfirm
        ? (0, exports.transformNestedAsset)(assetSchema, filteredResult)
        : (0, exports.transformAsset)(assetSchema, result);
};
exports.getParamsFromPrompt = getParamsFromPrompt;
const checkFileExtension = (filePath) => {
    const ext = path.extname(filePath);
    if (!ext || ext !== '.json') {
        throw new error_1.ValidationError(INVALID_JSON_FILE);
    }
};
exports.checkFileExtension = checkFileExtension;
const readParamsFile = (filePath) => {
    try {
        const params = fs.readFileSync(filePath, 'utf8');
        return params;
    }
    catch (err) {
        throw new error_1.ValidationError(FILE_NOT_FOUND);
    }
};
exports.readParamsFile = readParamsFile;
const getFileParams = (filePath) => {
    (0, exports.checkFileExtension)(filePath);
    return (0, exports.readParamsFile)(filePath);
};
exports.getFileParams = getFileParams;
//# sourceMappingURL=reader.js.map